///****************************   LICENSE   *****************************
/// Copyright (C) 2019, HUSEYIN TOKPUNAR http://huseyintokpinar.com/
/// Download & Update Latest Version: https://github.com/hhtokpinar/sqfEntity
/// Licensed under the Apache License, Version 2.0 (the "License");
/// you may not use this file except in compliance with the License.
/// You may obtain a copy of the License at

///    http://www.apache.org/licenses/LICENSE-2.0

/// Unless required by applicable law or agreed to in writing, software
/// distributed under the License is distributed on an "AS IS" BASIS,
/// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
/// See the License for the specific language governing permissions and
/// limitations under the License.
///**********************************************************************/

import 'package:analyzer/dart/constant/value.dart';


class SqfEntityModelBuilder extends SqfEntityModelBase {
  SqfEntityModelBuilder(this.model, this.instanceName);
  final DartObject model;
  final String instanceName;
//type 'String' is not a subtype of type 'SqfEntitySequenceBase' in type cast

  SqfEntityModelBase toModel() {
    final dbModel = _DbModel()
      ..modelName = getStringValue(model, 'modelName') ?? instanceName
      ..databaseName = getStringValue(model, 'databaseName')
      ..sequences = toSequenceList(getListValue(model, 'sequences'))
      ..databaseTables = toTableList(getListValue(model, 'databaseTables'))
      ..bundledDatabasePath = getStringValue(model, 'bundledDatabasePath');
    return dbModel;
  }

  SqfEntityTableBase toTable(DartObject obj) {
    return SqfEntityTableBase()
      ..tableName = getStringValue(
          obj, 'tableName') // getListValue(obj, 'fields').toString()
      ..primaryKeyName = getStringValue(obj, 'primaryKeyName')
      ..primaryKeyType = getTypeValue(obj, 'primaryKeyType') as PrimaryKeyType
      ..useSoftDeleting = getBoolValue(obj, 'useSoftDeleting')
      ..fields = toFields(getListValue(obj, 'fields'))
      ..defaultJsonUrl = getStringValue(obj, 'defaultJsonUrl')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntityTableBase> toTableList(List<DartObject> objTables) {
    print('SQFENTITY_GEN.DART: recognizing Tables');
    final sqfEntityTableList = List<SqfEntityTableBase>();
    for (var obj in objTables) {
      sqfEntityTableList.add(toTable(obj));
    }
    return sqfEntityTableList;
  }

  SqfEntitySequenceBase toSequence(DartObject obj) {
    return SqfEntitySequenceBase()
      ..sequenceName = getStringValue(obj, 'sequenceName')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntitySequenceBase> toSequenceList(List<DartObject> objSequences) {
    print('SQFENTITY_GEN.DART: recognizing Sequences');
    final sqfEntitySQEList = List<SqfEntitySequenceBase>();
    for (var obj in objSequences) {
      try {
        sqfEntitySQEList.add(toSequence(obj));
      } catch (e) {
        print(
            'SQFENTITY_GEN.DART: recognizing Sequences ERROR. tried obj=>${obj.toString()}\nError Message:${e.toString()}');
      }
    }
    return sqfEntitySQEList;
  }

  List<SqfEntityFieldType> toFields(List<DartObject> objFields) {
    final sqfEntityFieldList = List<SqfEntityFieldType>();
    for (var obj in objFields) {
      if (obj.toString().startsWith('SqfEntityFieldRelationship')) {
        sqfEntityFieldList.add(SqfEntityFieldRelationshipBase(
            toTable(obj.getField('parentTable')),
            getTypeValue(obj, 'deleteRule') as DeleteRule)
          ..defaultValue = getDynamicValue(obj, 'defaultValue'));
      } else {
        sqfEntityFieldList.add(SqfEntityFieldBase(
            getStringValue(obj, 'fieldName'),
            getTypeValue(obj, 'dbType') as DbType)
          ..defaultValue = getDynamicValue(obj, 'defaultValue')
          ..sequencedBy = obj
                  .getField('sequencedBy')
                  .toString()
                  .contains('SqfEntitySequenceAnnotation')
              ? toSequence(obj.getField('sequencedBy'))
              : null);
      }
    }
    return sqfEntityFieldList;
  }

  dynamic getDynamicValue(DartObject obj, String name) {
    final String type = obj
        .getField(name)
        .toString()
        .substring(0, obj.getField(name).toString().indexOf(' '))
        .toLowerCase();
    switch (type) {
      case 'string':
        return getStringValue(obj, name);
        break;
      case 'int':
        return getIntValue(obj, name);
        break;
      case 'bool':
        return getBoolValue(obj, name);
        break;
      default:
        return null;
    }
  }

  String getStringValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toStringValue() : null;
  bool getBoolValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toBoolValue() : null;
  int getIntValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toIntValue() : null;
  List<DartObject> getListValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toListValue() : null;
  dynamic getTypeValue(DartObject obj, String name) =>
      ifExist(obj, name) ? convertType(obj.getField(name)) : null;

  bool ifExist(DartObject obj, String name) => obj.getField(name) != null;
}

dynamic convertType(dynamic T) {
  final type = T.type;
  var types = Map<String, dynamic>();
  if (type.toString() == 'DbType') {
    types = dbTypes();
  } else if (type.toString() == 'PrimaryKeyType') {
    types = primaryKeyTypes();
  } else if (type.toString() == 'DeleteRule') {
    types = deleteRuleTypes();
  }
  for (var typ in types.entries) {
    if (T.toString().contains(typ.key)) {
      return typ.value;
    }
  }
  return type.toString();
}

Map<String, dynamic> dbTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DbType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] =
        typ; //.add(typ.toString().substring(typ.toString().indexOf('.') + 1));
  }
  return types;
}

Map<String, dynamic> primaryKeyTypes() {
  final types = Map<String, dynamic>();
  for (var typ in PrimaryKeyType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] =
        typ; //.add(typ.toString().substring(typ.toString().indexOf('.') + 1));
  }
  return types;
}

Map<String, dynamic> deleteRuleTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DeleteRule.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] =
        typ; //.add(typ.toString().substring(typ.toString().indexOf('.') + 1));
  }
  return types;
}


class SqfEntityConverter {
  SqfEntityConverter(this.m);
  final SqfEntityModelBase m;
  String _tableList() {
    final list = StringBuffer();
    for (final table in m.databaseTables) {
      list.writeln('Table${table.modelName}.getInstance,');
    }
    return list.toString();
  }

  String _tableListConst() {
    final list = StringBuffer();
    for (final table in m.databaseTables) {
      list.writeln('table${table.modelName},');
    }
    return list.toString();
  }

  String _sequenceList() {
    final list = StringBuffer();
    for (final seq in m.sequences) {
      list.writeln('Sequence${seq.modelName}.getInstance,');
    }
    return list.toString();
  }

  String get __createModelTables => _createModelTables();
  String get __createModelTablesConst => _createModelTablesConst();
  String get __createModelSequences => _createModelSequences();
  String get __tableList => _tableList();
  String get __tableListConst => _tableListConst();
  String get __sequenceList => _sequenceList();

  String createModelDatabase() {
    m.modelName = toModelName(m.modelName, 'MyDbModel');
    return '''
//  These classes was generated by SqfEntity 
//  Copyright (c) 2019, All rights reserved. Use of this source code is governed by a
//  Apache license that can be found in the LICENSE file.

//  To use these SqfEntity classes do following: 
//  - import model.dart into where to use
//  - start typing ex:${m.databaseTables.isNotEmpty ? m.databaseTables[0].modelName : ''}.select()... (add a few filters with fluent methods)...(add orderBy/orderBydesc if you want)...
//  - and then just put end of filters / or end of only select()  toSingle() / or toList() 
//  - you can select one or return List<yourObject> by your filters and orders
//  - also you can batch update or batch delete by using delete/update methods instead of tosingle/tolist methods
//    Enjoy.. Huseyin Tokpunar    

// BEGIN TABLES
$__createModelTables
// END TABLES

// BEGIN SEQUENCES
$__createModelSequences
// END SEQUENCES

// BEGIN DATABASE MODEL
class ${m.modelName} extends SqfEntityModelProvider {
  ${m.modelName}() {
    databaseName = '${m.databaseName}';
    databaseTables = [
      $__tableList
    ];
    // put defined sequence into the sequences list.
    sequences = [$__sequenceList];
    bundledDatabasePath =
        ${getStringNullableValue(m.bundledDatabasePath)}; //'assets/sample.db'; // This value is optional. When bundledDatabasePath is empty then EntityBase creats a new database when initializing the database
  }
}
// END DATABASE MODEL

  ''';
  }

  String createConstDatabase() {
    m.modelName = toModelName(m.modelName, 'MyDbModel');
    return '''
//  BEGIN ${m.databaseName} MODEL
//  Paste the following code into your model.dart
//  Enjoy.. Huseyin Tokpunar    

// BEGIN TABLES
$__createModelTablesConst
// END TABLES

// BEGIN DATABASE MODEL
@SqfEntityBuilder(myDbModel)
const ${tocamelCase(m.modelName)} = SqfEntityModel(
    modelName: '${m.modelName}',
    databaseName: '${m.databaseName}',
    databaseTables: [$__tableListConst],
    // put defined sequences into the sequences list.
    bundledDatabasePath:
        '${m.bundledDatabasePath}' 
);
// END ${m.databaseName} MODEL
''';
  }

  String _createModelTables() {
    final strTables = StringBuffer();

    for (final table in m.databaseTables) {
      strTables.writeln('''
// ${table.modelName} TABLE      
class Table${table.modelName} extends SqfEntityTableBase {
  Table${table.modelName}() {
    // declare properties of EntityTable
    tableName = '${table.tableName}';
    primaryKeyName = '${table.primaryKeyName}';
    primaryKeyType = ${table.primaryKeyType.toString()};
    useSoftDeleting = ${table.useSoftDeleting};
    // when useSoftDeleting is true, creates a field named 'isDeleted' on the table, and set to '1' this field when item deleted (does not hard delete)

    // declare fields
    fields = [
      ${_createModelFields(table)}
    ];
    super.init();
  }
  static SqfEntityTableBase _instance;
  static SqfEntityTableBase get getInstance {
    return _instance = _instance ?? Table${table.modelName}();
  }
}''');
    }

    return strTables.toString();
  }

  String _createModelTablesConst() {
    final strTables = StringBuffer();

    for (final table in m.databaseTables) {
      strTables.writeln('''
const table${toCamelCase(table.tableName)} = SqfEntityTable(
    tableName: '${table.tableName}',
    primaryKeyName: '${table.primaryKeyName}',
    primaryKeyType: ${table.primaryKeyType.toString()},
    useSoftDeleting: false,
    // when useSoftDeleting is true, creates a field named 'isDeleted' on the table, and set to '1' this field when item deleted (does not hard delete)
    fields: [
      ${_createModelFieldsConst(table)}
    ]);
}''');
    }

    return strTables.toString();
  }

  String _createModelSequences() {
    final strSequences = StringBuffer();
    for (var seq in m.sequences) {
      strSequences.writeln('''
// ${seq.sequenceName} SEQUENCE
class Sequence${seq.modelName} extends SqfEntitySequenceBase {
  Sequence${seq.modelName}() {
    sequenceName = '${seq.sequenceName}';
    maxValue = ${seq.maxValue};     /* optional. default is max int (9.223.372.036.854.775.807) */
    cycle = ${seq.cycle};      /* optional. default is false; */
    minValue = ${seq.minValue};    /* optional. default is 0 */
    incrementBy = ${seq.incrementBy}; /* optional. default is 1 */
    startWith = ${seq.startWith};  /* optional. default is 0 */
    super.init();
  }
  static Sequence${seq.modelName} _instance;
  static Sequence${seq.modelName} get getInstance {
    return _instance = _instance ?? Sequence${seq.modelName}();
  }
}''');
    }
    return strSequences.toString();
  }

  String _createModelFields(SqfEntityTableBase table) {
    final strFields = StringBuffer();

    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationshipBase) {
        strFields.writeln(
            'SqfEntityFieldRelationshipBase(Table${field.table.modelName}.getInstance, ${field.deleteRule.toString()}, defaultValue: ${getStringNullableValue(field.defaultValue)}),');
      } else {
        strFields.writeln(
            'SqfEntityFieldBase(\'${field.fieldName}\', ${field.dbType.toString()}, defaultValue: ${getStringNullableValue(field.defaultValue)}),');
      }
    }

    return strFields.toString();
  }

  String _createModelFieldsConst(SqfEntityTableBase table) {
    final strFields = StringBuffer();

    for (final field in table.fields) {
      strFields.writeln(
          'SqfEntityFieldBase(\'${field.fieldName}\', ${field.dbType.toString()}, defaultValue: null),');
    }
    return strFields.toString();
  }

  dynamic getStringNullableValue(dynamic value) {
    return value is String
        ? '\'${value.toString().replaceAll('\'', '\\\'')}\''
        : value;
  }

  /// Creates model of tables
  String createEntites() {
    //final String modelName = toString().replaceAll('Instance of', '').replaceAll("'", '').trim();
    final modelString = StringBuffer();
    if (m.databaseTables != null) {
      modelString.writeln('// BEGIN ENTITIES');
      for (var table in m.databaseTables) {
        table
          ..dbModel = m.modelName
          ..collections = getCollections(table);
        modelString
          ..writeln(SqfEntityObjectBuilder(table).toString())
          ..writeln(SqfEntityObjectField(table).toString())
          ..writeln(SqfEntityObjectFilterBuilder(table).toString())
          ..writeln(SqfEntityFieldBuilder(table).toString())
          ..writeln(SqfEntityObjectManagerBuilder(table).toString());
      }
    }

    if (m.sequences != null) {
      for (var sequence in m.sequences) {
        modelString.writeln(SqfEntitySequenceBuilder(sequence).toString());
      }
    }
    modelString..writeln('''class SequenceManager extends SqfEntityProvider {
SequenceManager() : super(${m.modelName}());
}''')..writeln('// END OF ENTITIES');
    print('SQFENTITIY: [databaseTables] Model was created successfully. ');
    return modelString.toString();
  }

  List<TableCollectionBase> getCollections(SqfEntityTableBase table) {
    final collectionList = List<TableCollectionBase>();
    for (var _table in m.databaseTables) {
      for (var field
          in _table.fields.where((f) => f is SqfEntityFieldRelationshipBase)) {
        if (field.table.tableName == table.tableName) {
          collectionList.add(TableCollectionBase(
              _table, field as SqfEntityFieldRelationshipBase));
        }
      }
    }
    return collectionList;
  }
}


// BEGIN ANNOTATIONS

class SqfEntityBuilder {
  const SqfEntityBuilder(this.model);
  final SqfEntityModel model;
}

class SqfEntityTable {
  const SqfEntityTable({
    this.tableName,
    this.primaryKeyName,
    this.fields,
    this.useSoftDeleting,
    this.primaryKeyType,
    this.defaultJsonUrl,
    this.modelName,
  });
  final String tableName;
  final String primaryKeyName;
  final List<SqfEntityField> fields;
  //List<TableCollection> collections;
  final bool useSoftDeleting;
  //bool primaryKeyisIdentity = true;
  final PrimaryKeyType primaryKeyType;
  final String modelName;
  final String defaultJsonUrl;
  //final String dbModel;
}

class TableCollection {
  const TableCollection(this.childTable, this.childTableField);
  final SqfEntityTable childTable;
  final SqfEntityFieldRelationship childTableField;
}

class SqfEntitySequence {
  const SqfEntitySequence(
      {this.sequenceName,
      this.startWith,
      this.incrementBy,
      this.minValue,
      this.maxValue,
      this.cycle,
      this.modelName});

  /// Name of the sequence.
  final String sequenceName;

  /// starting value from where the sequence starts. startWith value should be greater than or equal
  /// to minimum value and less than equal to maximum value.
  final int startWith;

  ///Value by which sequence will increment itself. Increment_value can be positive or negative.
  final int incrementBy;

  /// Minimum value of the sequence.
  final int minValue;

  /// Maximum value of the sequence.
  final int maxValue;

  /// cycle if true then, When sequence reaches its max value it starts from beginning. otherwise An exception will be thrown if sequence exceeds its max_value.
  final bool cycle;

  final String modelName;
}

class SqfEntityField {
  const SqfEntityField(this.fieldName, this.dbType,
      {this.defaultValue, this.sequencedBy});
  final String fieldName;
  final DbType dbType;
  final dynamic defaultValue;
  final SqfEntitySequence sequencedBy;
}

class SqfEntityFieldRelationship implements SqfEntityField {
  const SqfEntityFieldRelationship(
      {this.parentTable, this.deleteRule, this.fieldName, this.defaultValue});
  @override
  final String fieldName;
  @override
  final dynamic defaultValue;
  final DeleteRule deleteRule;
  final SqfEntityTable parentTable;
  @override
  SqfEntitySequence get sequencedBy => null;
  @override
  DbType get dbType => null;
}

class SqfEntityModel {
  const SqfEntityModel(
      {this.databaseName,
      this.databaseTables,
      this.bundledDatabasePath,
      this.modelName,
      this.sequences});
  // STEPS FOR CREATE YOUR DB CONTEXT

  // 1. declare your sqlite database name
  final String databaseName;

  // This value is optional. When bundledDatabasePath is empty then EntityBase creats a new database when initializing database
  final String bundledDatabasePath;

  // 2. Add the object you defined above to your list of database tables and sequences.
  final List<SqfEntityTable> databaseTables;
  final List<SqfEntitySequence> sequences;

  // set this property to your DBModel.dart path for ex: "import 'MyDbModel.dart';"
  final String modelName;
  // that's all.. one more step left for create models.dart file.
  // ATTENTION: Defining the table here provides automatic processing for database configuration only.
  // you may call the SqfEntityDbContext.createModel(MyDbModel.databaseTables) function to create your model and use it in your project
}

// END ANNOTATIONS

// BEGIN CONVERTERS
class ConvertedModel extends SqfEntityModelBase {}

class SqfEntityModelConverter {
  SqfEntityModelConverter(this.model);
  final SqfEntityModel model;

  SqfEntityModelBase toModelBase() {
    return ConvertedModel()
      ..databaseName = model.databaseName
      ..modelName = model.modelName
      ..databaseTables = toTables()
      ..sequences = toSequences()
      ..bundledDatabasePath = model.bundledDatabasePath;
  }

  List<SqfEntityTableBase> toTables() {
    if (model.databaseTables == null) return null;
    final tables = List<SqfEntityTableBase>();
    for (final table in model.databaseTables) {
      tables.add(toTable(table));
    }
    return tables;
  }

  SqfEntityTableBase toTable(SqfEntityTable table) {
    return SqfEntityTableBase()
      ..tableName = table.tableName
      ..modelName = table.modelName
      ..primaryKeyName = table.primaryKeyName
      ..primaryKeyType = table.primaryKeyType
      ..useSoftDeleting = table.useSoftDeleting
      ..defaultJsonUrl = table.defaultJsonUrl
      ..fields = toFields(table)
      ..init();
  }

  List<SqfEntityFieldType> toFields(SqfEntityTable table) {
    if (table.fields == null) return null;
    final fields = List<SqfEntityFieldType>();
    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationship) {
        fields.add(SqfEntityFieldRelationshipBase(
            toTable(field.parentTable), field.deleteRule)
          ..defaultValue = field.defaultValue);
      } else {
        fields.add(SqfEntityFieldBase(field.fieldName, field.dbType)
          ..defaultValue = field.defaultValue
          ..sequencedBy =
              field.sequencedBy == null ? null : toSequence(field.sequencedBy));
      }
    }
    return fields;
  }

  List<SqfEntitySequenceBase> toSequences() {
    if (model.sequences == null) return null;
    final sequenceList = List<SqfEntitySequenceBase>();
    for (final seq in model.sequences) {
      sequenceList.add(toSequence(seq));
    }
    return sequenceList;
  }

  SqfEntitySequenceBase toSequence(SqfEntitySequence seq) {
    if (seq == null) {
      return null;
    } else {
      return SqfEntitySequenceBase()
        ..sequenceName = seq.sequenceName
        ..modelName = seq.modelName
        ..minValue = seq.minValue
        ..maxValue = seq.maxValue
        ..startWith = seq.startWith
        ..incrementBy = seq.incrementBy
        ..cycle = seq.cycle
        ..init();
    }
  }
}
// END CONVERTERS

// BEGIN SQFENTITY BUILDERS

class SqfEntityFieldBuilder {
  SqfEntityFieldBuilder(this._table);
  SqfEntityTableBase _table;
  String get _createObjectField => __createObjectField();

  @override
  String toString() {
    final String toString = '''\n
// region ${_table.modelName}Fields
class ${_table.modelName}Fields {
$_createObjectField
}
// endregion ${_table.modelName}Fields
''';

    return toString;
  }

  String __createObjectField() {
    final retVal = StringBuffer(
        '''static TableField _f${toCamelCase(_table.primaryKeyName)};
  static TableField get ${_table.primaryKeyName.toLowerCase()} {
  return _f${toCamelCase(_table.primaryKeyName)} = _f${toCamelCase(_table.primaryKeyName)} ?? SqlSyntax.setField(_f${toCamelCase(_table.primaryKeyName)}, '${_table.primaryKeyName.toLowerCase()}', ${DbType.integer.toString()});
  }
''');

    for (SqfEntityFieldType field in _table.fields) {
      retVal.writeln(
          '''static TableField _f${field.fieldName.substring(0, 1).toUpperCase() + field.fieldName.substring(1)};
  static TableField get ${field.fieldName} {
   return _f${field.fieldName.substring(0, 1).toUpperCase() + field.fieldName.substring(1)} = _f${field.fieldName.substring(0, 1).toUpperCase() + field.fieldName.substring(1)} ?? SqlSyntax.setField(_f${field.fieldName.substring(0, 1).toUpperCase() + field.fieldName.substring(1)}, '${field.fieldName}', ${field.dbType.toString()});
  }
''');
    }
    if (_table.useSoftDeleting) {
      retVal.writeln('''static TableField _fIsDeleted;
  static TableField get isDeleted {
   return  _fIsDeleted = _fIsDeleted ?? SqlSyntax.setField(_fIsDeleted, 'isDeleted', ${DbType.integer.toString()});
  }''');
    }
    return retVal.toString();
  }
}

class SqfEntityObjectBuilder {
  SqfEntityObjectBuilder(this._table);
  SqfEntityTableBase _table;
  String get _createProperties => __createProperties();
  String get _createObjectRelations => __createObjectRelations();
  String get _createObjectCollections => __createObjectCollections();
  String get _createObjectCollectionsToMap => __createObjectCollectionsToMap();
  String get _createConstructure => __createConstructure();
  String get _createConstructureArgs => __createConstructureArgs();
  String get _toMapString => __toMapString();
  String get _toFromMapString => __toFromMapString();
  String get _fromWebUrl => __fromWebUrl();
  String get _saveMethod => __saveMethod();
  String get _upsertMethod => __upsertMethod();
  String get _deleteMethodSingle => __deleteMethodSingle();
  String get _recoveryMethodSingle => __recoveryMethodSingle();
  String get _createDefaultValues => __createDefaultValues();
  @override
  String toString() {
    final String toString = '''
// region ${_table.modelName}
class ${_table.modelName} {
  ${_table.modelName}({this.${_table.primaryKeyName}, $_createConstructure}) { setDefaultValues();}
  ${_table.modelName}.withFields($_createConstructure){ setDefaultValues();}
  ${_table.modelName}.withId(this.${_table.primaryKeyName}, $_createConstructure){ setDefaultValues();}
  ${_table.modelName}.fromMap(Map<String, dynamic> o) {
    $_toFromMapString
    }
  // FIELDS
  $_createProperties      // end FIELDS
  $_createObjectRelations
  $_createObjectCollections
  static const bool _softDeleteActivated=${_table.useSoftDeleting.toString()};
  ${_table.modelName}Manager __mn${_table.modelName};

  ${_table.modelName}Manager get _mn${_table.modelName} {
    return __mn${_table.modelName} = __mn${_table.modelName} ?? ${_table.modelName}Manager();
  }

  // methods
  Map<String, dynamic> toMap({bool forQuery=false}){
    final map = Map<String, dynamic>();
    $_toMapString
    return map;
    }

  // methods
  Future<Map<String, dynamic>> toMapWithChilds() async{
    final forQuery = false;
    final map = Map<String, dynamic>();
    $_toMapString
    $_createObjectCollectionsToMap
    return map;
    }

  /// This method always returns Json String
  String toJson() {
    return json.encode(toMap());
  }
  
  /// This method always returns Json String
  Future<String> toJsonWithChilds() async {
    return json.encode(await toMapWithChilds());
  }


  List<dynamic> toArgs() {
    return[${_table.primaryKeyName},${_createConstructure.replaceAll('this.', '')}];   
  }  

  $_fromWebUrl  
  static Future<List<${_table.modelName}>> fromWebUrl(String url) async {
    try {
      final response = await http.get(url);
      return await fromJson(response.body);
    } catch (e) {
      print('SQFENTITY ERROR ${_table.modelName}.fromWebUrl: ErrorMessage: \${e.toString()}');
      return null;
    }
  }

  static Future<List<${_table.modelName}>> fromJson(String jsonBody) async{
    final Iterable list = await json.decode(jsonBody) as Iterable;
    var objList = List<${_table.modelName}>();
    try {
      objList = list.map((${_table._modelLowerCase}) => ${_table.modelName}.fromMap(${_table._modelLowerCase} as Map<String, dynamic>)).toList();
    } catch (e) {
      print('SQFENTITY ERROR ${_table.modelName}.fromJson: ErrorMessage: \${e.toString()}');
    }
    return objList;
  }

  static Future<List<${_table.modelName}>> fromObjectList(Future<List<dynamic>> o) async {
    final ${_table._modelLowerCase}sList = List<${_table.modelName}>();
    final data = await o;
      for (int i = 0; i < data.length; i++) {
        ${_table._modelLowerCase}sList.add(${_table.modelName}.fromMap(data[i] as Map<String, dynamic>));
      }
    return ${_table._modelLowerCase}sList;
  }

  static List<${_table.modelName}> fromMapList(List<Map<String, dynamic>> query) {
    final List<${_table.modelName}> ${_table._modelLowerCase}s = List<${_table.modelName}>();
    for (Map map in query) {
      ${_table._modelLowerCase}s.add(${_table.modelName}.fromMap(map as Map<String, dynamic>));
    }
    return ${_table._modelLowerCase}s;
  }


  /// returns ${_table.modelName} by ID if exist, otherwise returns null
  /// <param name='${_table.primaryKeyName}'>Primary Key Value</param>
  /// <returns>returns ${_table.modelName} if exist, otherwise returns null</returns>
  Future<${_table.modelName}> getById(${(_table.primaryKeyType == PrimaryKeyType.text ? 'String' : 'int')} ${_table.primaryKeyName.toLowerCase()}) async{
    ${_table.modelName} ${_table._modelLowerCase}Obj;
    final data = await _mn${_table.modelName}.getById(id);
    if (data.length != 0) 
        {${_table._modelLowerCase}Obj = ${_table.modelName}.fromMap(data[0] as Map<String, dynamic>);}
    else
        {${_table._modelLowerCase}Obj = null;}
    return ${_table._modelLowerCase}Obj;
  }

  $_saveMethod

  /// <summary>
  /// saveAll method saves the sent List<${_table.modelName}> as a batch in one transaction 
  /// </summary>
  /// <returns> Returns a <List<BoolResult>> </returns>
  Future<List<BoolResult>> saveAll(List<${_table.modelName}> ${toPluralName(_table._modelLowerCase)}) async {
    final results = _mn${_table.modelName}.saveAll('INSERT OR REPLACE INTO ${_table.tableName} (${_table.primaryKeyName}, ${_createConstructure.replaceAll('this.', '')})  VALUES (?,$_createConstructureArgs)',${toPluralName(_table._modelLowerCase)});
    return results;
  }
  $_upsertMethod
  /// <summary>
  /// inserts or replaces the sent List<Todo> as a batch in one transaction.
  /// upsertAll() method is faster then saveAll() method. upsertAll() should be used when you are sure that the primary key is greater than zero
  /// </summary>
  /// <returns> Returns a <List<BoolResult>> </returns>
  Future<List<BoolResult>> upsertAll(List<${_table.modelName}> ${toPluralName(_table._modelLowerCase)}) async {
    final results = await _mn${_table.modelName}.rawInsertAll(
        'INSERT OR REPLACE INTO ${_table.tableName} (${_table.primaryKeyName}, ${_createConstructure.replaceAll('this.', '')})  VALUES (?,$_createConstructureArgs)', ${toPluralName(_table._modelLowerCase)});
    return results;
  }


  /// <summary>
  /// Deletes ${_table.modelName}
  /// </summary>
  /// <returns>BoolResult res.success=Deleted, not res.success=Can not deleted</returns>
  Future<BoolResult> delete([bool hardDelete=false]) async {
    print('SQFENTITIY: delete ${_table.modelName} invoked (${_table.primaryKeyName}=\$${_table.primaryKeyName})');
    $_deleteMethodSingle
  }
    $_recoveryMethodSingle
  //private ${_table.modelName}FilterBuilder _Select;
  ${_table.modelName}FilterBuilder select(
      {List<String> columnsToSelect, bool getIsDeleted}) {
    return ${_table.modelName}FilterBuilder(this)
    .._getIsDeleted = getIsDeleted==true
    ..qparams.selectColumns = columnsToSelect;
  }

  ${_table.modelName}FilterBuilder distinct(
      {List<String> columnsToSelect, bool getIsDeleted}) {
    return ${_table.modelName}FilterBuilder(this)
    .._getIsDeleted = getIsDeleted==true
    ..qparams.selectColumns = columnsToSelect
    ..qparams.distinct = true;
  }

  void setDefaultValues() {
    $_createDefaultValues
  }
  //end methods
}
// endregion ${_table._modelLowerCase}

            ''';
    return toString;
  }

  String __createProperties() {
    final _retVal = StringBuffer(
        '${_table.primaryKeyType == PrimaryKeyType.text ? 'String' : 'int'} ${_table.primaryKeyName};');
    for (final field in _table.fields) {
      _retVal.writeln('    ${field.toPropertiesString()};');
    }
    if (_table.useSoftDeleting) {
      _retVal.writeln('    bool isDeleted;');
    }

    return _retVal.toString();
  }

  String __createConstructure() {
    final _retVal = StringBuffer();
    for (final field in _table.fields) {
      _retVal.write(', ${field.toConstructureString()}');
    }
    if (_table.useSoftDeleting) {
      _retVal.write(',this.isDeleted');
    }
    return _retVal.toString().substring(1);
  }

  String __createConstructureArgs() {
    final _retVal = StringBuffer();
    for (var i = 0; i < _table.fields.length; i++) {
      _retVal.write(',?');
    }
    if (_table.useSoftDeleting) {
      _retVal.write(',?');
    }
    return _retVal.toString().substring(1);
  }

  String __toMapString() {
    final _retVal = StringBuffer(
        'if (${_table.primaryKeyName} != null) {map[\'${_table.primaryKeyName}\'] = ${_table.primaryKeyName};}');
    for (var field in _table.fields) {
      _retVal.writeln('    ${field.toMapString()}');
    }
    if (_table.useSoftDeleting) {
      _retVal.writeln(
          '  if (isDeleted != null) {map[\'isDeleted\'] = forQuery? (isDeleted ? 1 : 0):isDeleted;}');
    }
    return _retVal.toString();
  }

  String __toFromMapString() {
    final _retVal = StringBuffer(
        '${_table.primaryKeyName} = o[\'${_table.primaryKeyName}\'] as ${_table.primaryKeyType == PrimaryKeyType.text ? 'String' : 'int'};');
    for (final field in _table.fields) {
      _retVal.writeln('    ${field.toFromMapString()}');
    }
    if (_table.useSoftDeleting) {
      _retVal.writeln(
          '  isDeleted = o[\'isDeleted\'] != null ? o[\'isDeleted\'] == 1 : null;');
    }
    return _retVal.toString();
  }

  String __createDefaultValues() {
    final _retVal = StringBuffer();
    for (final field in _table.fields.where((f) => f.defaultValue != null)) {
      switch (field.dbType) {
        case DbType.text:
          field.defaultValue =
              "'${field.defaultValue.replaceAll('\'', '\\\'')}'";
          break;
        case DbType.bool:
          field.defaultValue =
              '${field.defaultValue == '1' || field.defaultValue == 'true' ? 'true' : 'false'}';
          break;
        default:
      }
      _retVal.writeln(
          ' ${field.fieldName}=${field.fieldName} ?? ${field.defaultValue};');
    }
    if (_table.useSoftDeleting) {
      _retVal.writeln(' isDeleted= isDeleted ?? false;');
    }
    return _retVal.toString();
  }

  String __createObjectRelations() {
    var retVal = '';

    for (final field
        in _table.fields.where((f) => f is SqfEntityFieldRelationshipBase)) {
      if (field is SqfEntityFieldRelationshipBase) {
        retVal += '''
       Future<${field.relationshipName}> get${field.relationshipName}([VoidCallback ${field.relationshipName.toLowerCase()}(${field.table.modelName} o)]) async{
          final obj = await ${field.relationshipName}().getById(${field.fieldName});
          if(${field.relationshipName.toLowerCase()} != null){
             ${field.relationshipName.toLowerCase()}(obj);
            }
          return obj;
        }
      ''';
      }
    }
    if (retVal.isNotEmpty) {
      retVal = '\n// RELATIONSHIPS\n$retVal// END RELATIONSHIPS\n';
    }
    return retVal;
  }

  String __createObjectCollections() {
    final retVal = StringBuffer();
    for (var tableCollecion in _table.collections) {
      retVal.writeln(
          '''${tableCollecion.childTable.modelName}FilterBuilder get${toPluralName(tableCollecion.childTable.modelName)}({List<String> columnsToSelect, bool getIsDeleted}){
         return ${tableCollecion.childTable.modelName}().select(columnsToSelect: columnsToSelect,getIsDeleted: getIsDeleted).${tableCollecion.childTableField.fieldName}.equals(${tableCollecion.childTableField.table.primaryKeyName}).and;
        }''');
    }
    return retVal.length > 0
        ? '\n// COLLECTIONS\n${retVal.toString()}// END COLLECTIONS\n'
        : '';
  }

  String __createObjectCollectionsToMap() {
    final retVal = StringBuffer();
    for (var tableCollecion in _table.collections) {
      retVal.writeln(
          '''if(!forQuery) {map['${toPluralName(tableCollecion.childTable.tableName)}'] = await get${toPluralName(tableCollecion.childTable.modelName)}().toMapList();}''');
    }
    return retVal.length > 0
        ? '\n// COLLECTIONS\n${retVal.toString()}// END COLLECTIONS\n'
        : '';
  }

  String __fromWebUrl() {
    if (_table.defaultJsonUrl != null && _table.defaultJsonUrl.isNotEmpty) {
      return '''
       static Future<List<${_table.modelName}>> fromWeb([VoidCallback ${_table._modelLowerCase}List(List<${_table.modelName}> o)]) async {
            final objList = await fromWebUrl('${_table.defaultJsonUrl}');
            if (${_table._modelLowerCase}List != null) {${_table._modelLowerCase}List (objList); } 
            return objList;
        }
            ''';
    } else {
      return '';
    }
  }

  String __deleteMethodSingle() {
    var retVal = '';
    final varResult = 'var result= BoolResult();';

    for (var tableCollecion in _table.collections) {
      switch (tableCollecion.childTableField.deleteRule) {
        case DeleteRule.SET_NULL:
          retVal += '''
    {result = await ${tableCollecion.childTable.modelName}().select().${tableCollecion.childTableField.fieldName}.equals(${tableCollecion.childTableField.table.primaryKeyName}).update({"${tableCollecion.childTableField.fieldName}": null});}
    if (!result.success) {return result;}
    else''';
          break;
        case DeleteRule.NO_ACTION:
          retVal += '''
    if (await ${tableCollecion.childTable.modelName}().select().${tableCollecion.childTableField.fieldName}.equals(${tableCollecion.childTableField.table.primaryKeyName}).toCount()>0) return BoolResult(success: false,errorMessage: "SQFENTITY ERROR: The DELETE statement conflicted with the REFERENCE RELATIONSHIP '${tableCollecion.childTable.modelName}.${tableCollecion.childTableField.fieldName}'");
    else''';

          break;
        case DeleteRule.CASCADE:
          retVal += '''
    {result = await ${tableCollecion.childTable.modelName}().select().${tableCollecion.childTableField.fieldName}.equals(${tableCollecion.childTableField.table.primaryKeyName}).delete(hardDelete);}
    if (!result.success) {return result;}
          else
           ''';
          break;
        case DeleteRule.SET_DEFAULT_VALUE:
          retVal += '''
    {result = await ${tableCollecion.childTable.modelName}().select().${tableCollecion.childTableField.fieldName}.equals(${tableCollecion.childTableField.table.primaryKeyName}).update({"${tableCollecion.childTableField.fieldName}": ${tableCollecion.childTableField.defaultValue}});}
    if (!result.success) {return result;}
    else''';
          break;
        default:
      }
    }
    if (retVal != '') {
      retVal = varResult + retVal;
    }

    return retVal += '''
    if (!_softDeleteActivated || hardDelete${_table.useSoftDeleting ? ' || isDeleted' : ''}) {
    return _mn${_table.modelName}.delete(QueryParams(whereString: '${_table.primaryKeyName}=\$${_table.primaryKeyName}'));}
    else {
    return _mn${_table.modelName}.updateBatch(QueryParams(whereString: '${_table.primaryKeyName}=\$${_table.primaryKeyName}'), {'isDeleted': 1});}''';
  }

  String __recoveryMethodSingle() {
    if (!_table.useSoftDeleting) {
      return '';
    }
    var retVal = '';
    final varResult = 'var result= BoolResult();';

    for (var tableCollecion in _table.collections) {
      switch (tableCollecion.childTableField.deleteRule) {
        case DeleteRule.SET_NULL:
          // IN THIS CASE YOU CAN NOT RECOVER CHILD ROWS
          break;
        case DeleteRule.CASCADE:
          if (tableCollecion.childTable.useSoftDeleting) {
            retVal += '''
    if(recoverChilds) {result = await ${tableCollecion.childTable.modelName}().select(getIsDeleted: true).isDeleted.equals(true).and.${tableCollecion.childTableField.fieldName}.equals(${tableCollecion.childTableField.table.primaryKeyName}).update({'isDeleted': 0});}
    if (!result.success && recoverChilds) {return result;}
    else''';
          }
          break;
        case DeleteRule.SET_DEFAULT_VALUE:
          //IN THIS CASE YOU CAN NOT RECOVER CHILD ROWS
          break;
        default:
      }
    }
    if (retVal != '') {
      retVal = varResult + retVal;
    }
    retVal += '''
    {return _mn${_table.modelName}.updateBatch(QueryParams(whereString: '${_table.primaryKeyName}=\$${_table.primaryKeyName}'), {'isDeleted': 0});}''';

    return '''
    /// <summary>
    /// Recover ${_table.modelName}>
    /// </summary>
    /// <returns>BoolResult res.success=Recovered, not res.success=Can not recovered</returns>
    Future<BoolResult> recover([bool recoverChilds=true]) async {
    print('SQFENTITIY: recover ${_table.modelName} invoked (${_table.primaryKeyName}=\$${_table.primaryKeyName})');
        $retVal 
    }''';
  }

  String __upsertMethod() {
    return _table.primaryKeyType == PrimaryKeyType.text
        ? ''
        : '''
  /// <summary>
  /// Updates if the record exists, otherwise adds a new row
  /// </summary>
  /// <returns>Returns ${_table.primaryKeyName}</returns>
  Future<int> _upsert() async {
    return ${_table.primaryKeyName} = await _mn${_table.modelName}.rawInsert(
        'INSERT OR REPLACE INTO ${_table.tableName} (${_table.primaryKeyName}, ${_createConstructure.replaceAll('this.', '')})  VALUES (?,$_createConstructureArgs)', [${_table.primaryKeyName},${_createConstructure.replaceAll('this.', '')}]);
  }
''';
  }

  String __saveMethod() {
    final seq = StringBuffer();
    for (var sequencedField
        in _table.fields.where((i) => i.sequencedBy != null)) {
      seq.writeln(
          'if (${_table.primaryKeyName} != null) { ${sequencedField.fieldName} = await ${sequencedField.sequencedBy.modelName}().nextVal(); save();}');
    }
    var retVal = '';
    if (_table.primaryKeyType == PrimaryKeyType.text) {
      retVal = '''
  /// <summary>
  /// Saves the object. If the Primary Key (${_table.primaryKeyName}) field is null, returns Error. 
  /// INSERTS (If not exist) OR REPLACES (If exist) data while Primary Key is not null. 
  /// Call the saveAs() method if you do not want to save it when there is another row with the same ${_table.primaryKeyName}
  /// </summary>
  /// <returns>Returns BoolResult</returns>
  Future<BoolResult> save() async {
    final result = BoolResult(success: false);
    try {         
      await _mn${_table.modelName}.rawInsert(
        "INSERT OR REPLACE INTO ${_table.tableName} (${_table.primaryKeyName}, ${_createConstructure.replaceAll("this.", "")})  VALUES (?,$_createConstructureArgs)", [${_table.primaryKeyName},${_createConstructure.replaceAll("this.", "")}]);
      result.success=true;
    } catch (e){
      result.errorMessage = e.toString();
    }
    return result;
  }
  /// <summary>
  /// saveAs ${_table.modelName}
  /// Use this method if you do not want to update existing row when conflicts another row that have the same ${_table.primaryKeyName}
  /// </summary>
  /// Returns a BoolResult
  Future<BoolResult> saveAs() async {
    final result = BoolResult(success: false);
    try {         
      await _mn${_table.modelName}.rawInsert(
        "INSERT INTO ${_table.tableName} (${_table.primaryKeyName}, ${_createConstructure.replaceAll("this.", "")})  VALUES (?,$_createConstructureArgs)", [${_table.primaryKeyName},${_createConstructure.replaceAll("this.", "")}]);
    ${seq.toString()}              
      result.success=true;
    } catch (e){
      result.errorMessage = e.toString();
    }
    return result;
  }
       
  ''';
    } else {
      retVal = '''
  /// <summary>
  /// Saves the object. If the ${_table.primaryKeyName} field is null, saves as a new record and returns new ${_table.primaryKeyName}, if ${_table.primaryKeyName} is not null then updates record
  /// </summary>
  /// <returns>Returns ${_table.primaryKeyName}</returns>
  Future<int> save() async {
    if (${_table.primaryKeyName} == null || ${_table.primaryKeyName} == 0) {
      ${_table.primaryKeyName} = await _mn${_table.modelName}.insert(this);
      ${seq.toString()}
        }
    else {
      ${_table.primaryKeyName}= await _upsert(); }
    return ${_table.primaryKeyName};
  }

  /// <summary>
  /// saveAs ${_table.modelName}. Returns a new Primary Key value of ${_table.modelName}
  /// </summary>
  /// <returns>Returns a new Primary Key value of ${_table.modelName}</returns>
  Future<int> saveAs() async {
    ${_table.primaryKeyName} = null;
    return save();
  }
  
  ''';
    }

    return retVal;
  }
}

class SqfEntityObjectManagerBuilder {
  SqfEntityObjectManagerBuilder(this._table);
  SqfEntityTableBase _table;

  @override
  String toString() {
    final String toString = '''//region ${_table.modelName}Manager
class ${_table.modelName}Manager extends SqfEntityProvider {
  ${_table.modelName}Manager():super(${_table.dbModel}(),tableName: _tableName, colId: _colId);
  static String _tableName = '${_table.tableName}';
  static String _colId = '${_table.primaryKeyName}';
 
}
//endregion ${_table.modelName}Manager''';

    return toString;
  }
}

class SqfEntitySequenceBuilder {
  SqfEntitySequenceBuilder(this.sequence);
  SqfEntitySequenceBase sequence;
  @override
  String toString() {
    return '''
/// Region SEQUENCE ${sequence.modelName}    
class ${sequence.modelName} {
  /// Assigns a new value when it is triggered and returns the new value
  /// returns Future<int>
  Future<int> nextVal([VoidCallback nextval(int o)]) async {
    final val = await SequenceManager().sequence(Sequence${sequence.modelName}.getInstance, true);
    if (nextval != null) {
      nextval(val);
    }
    return val;
  }
  /// Get the current value
  /// returns Future<int>
  Future<int> currentVal([VoidCallback currentval(int o)]) async {
    final val = await SequenceManager().sequence(Sequence${sequence.modelName}.getInstance, false);
    if (currentval != null) {
      currentval(val);
    }
    return val;
  }
  /// Reset sequence to start value
  /// returns start value
  Future<int> reset([VoidCallback currentval(int o)]) async {
    final val = await SequenceManager().sequence(Sequence${sequence.modelName}.getInstance, false, reset: true);
    if (currentval != null) {
      currentval(val);
    }
    return val;
  }
}
/// End Region SEQUENCE ${sequence.modelName}
''';
  }
}

class SqfEntityObjectField {
  SqfEntityObjectField(this._table);
  SqfEntityTableBase _table;
  @override
  String toString() {
    final String toString = '''// region ${_table.modelName}Field
class ${_table.modelName}Field extends SearchCriteria {
  ${_table.modelName}Field(this.${_table._modelLowerCase}FB) {
    param = DbParameter();
  }
  DbParameter param;
  String _waitingNot = '';
  ${_table.modelName}FilterBuilder ${_table._modelLowerCase}FB;
  
  ${_table.modelName}Field get not {
    _waitingNot = ' NOT ';
    return this;
  }

  ${_table.modelName}FilterBuilder equals(var pValue) {
    param.expression = '=';
    ${_table._modelLowerCase}FB._addedBlocks = _waitingNot == ''
        ? setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param, SqlSyntax.EQuals,
            ${_table._modelLowerCase}FB._addedBlocks)
        : setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param, SqlSyntax.NotEQuals,
            ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder isNull() {
    ${_table._modelLowerCase}FB._addedBlocks = setCriteria(
        0,
        ${_table._modelLowerCase}FB.parameters,
        param,
        SqlSyntax.IsNULL.replaceAll(SqlSyntax.notKeyword, _waitingNot),
        ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder contains(dynamic pValue) {
     if(pValue != null){${_table._modelLowerCase}FB._addedBlocks = setCriteria(
        '%\${pValue.toString()}%',
        ${_table._modelLowerCase}FB.parameters,
        param,
        SqlSyntax.Contains.replaceAll(SqlSyntax.notKeyword, _waitingNot),
        ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;}
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder startsWith(dynamic pValue) {
     if(pValue != null){${_table._modelLowerCase}FB._addedBlocks = setCriteria(
        '\${pValue.toString()}%',
        ${_table._modelLowerCase}FB.parameters,
        param,
        SqlSyntax.Contains.replaceAll(SqlSyntax.notKeyword, _waitingNot),
        ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;}
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder endsWith(dynamic pValue) {
     if(pValue != null){${_table._modelLowerCase}FB._addedBlocks = setCriteria(
        '%\${pValue.toString()}',
        ${_table._modelLowerCase}FB.parameters,
        param,
        SqlSyntax.Contains.replaceAll(SqlSyntax.notKeyword, _waitingNot),
        ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;}
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder between(dynamic pFirst, dynamic pLast) {
    if (pFirst != null && pLast != null) {
      ${_table._modelLowerCase}FB._addedBlocks = setCriteria(
          pFirst,
          ${_table._modelLowerCase}FB.parameters,
          param,
          SqlSyntax.Between.replaceAll(SqlSyntax.notKeyword, _waitingNot),
          ${_table._modelLowerCase}FB._addedBlocks,
          pLast);
    } else if (pFirst != null) {
      if (_waitingNot != '') {
        ${_table._modelLowerCase}FB._addedBlocks = setCriteria(pFirst, ${_table._modelLowerCase}FB.parameters,
            param, SqlSyntax.LessThan, ${_table._modelLowerCase}FB._addedBlocks); }
      else {
        ${_table._modelLowerCase}FB._addedBlocks = setCriteria(pFirst, ${_table._modelLowerCase}FB.parameters,
            param, SqlSyntax.GreaterThanOrEquals, ${_table._modelLowerCase}FB._addedBlocks); }
    } else if (pLast != null) {
      if (_waitingNot != '') {
        ${_table._modelLowerCase}FB._addedBlocks = setCriteria(pLast, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.GreaterThan, ${_table._modelLowerCase}FB._addedBlocks);}
      else {
        ${_table._modelLowerCase}FB._addedBlocks = setCriteria(pLast, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.LessThanOrEquals, ${_table._modelLowerCase}FB._addedBlocks); }
    }
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder greaterThan(dynamic pValue) {
    param.expression = '>';
    ${_table._modelLowerCase}FB._addedBlocks = _waitingNot == ''
        ? setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.GreaterThan, ${_table._modelLowerCase}FB._addedBlocks)
        : setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.LessThanOrEquals, ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder lessThan(dynamic pValue) {
    param.expression = '<';
    ${_table._modelLowerCase}FB._addedBlocks = _waitingNot == ''
        ? setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param, SqlSyntax.LessThan,
            ${_table._modelLowerCase}FB._addedBlocks)
        : setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.GreaterThanOrEquals, ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder greaterThanOrEquals(dynamic pValue) {
    param.expression = '>=';
    ${_table._modelLowerCase}FB._addedBlocks = _waitingNot == ''
        ? setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.GreaterThanOrEquals, ${_table._modelLowerCase}FB._addedBlocks)
        : setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param, SqlSyntax.LessThan,
            ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder lessThanOrEquals(dynamic pValue) {
    param.expression = '<=';
    ${_table._modelLowerCase}FB._addedBlocks = _waitingNot == ''
        ? setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.LessThanOrEquals, ${_table._modelLowerCase}FB._addedBlocks)
        : setCriteria(pValue, ${_table._modelLowerCase}FB.parameters, param,
            SqlSyntax.GreaterThan, ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }

  ${_table.modelName}FilterBuilder inValues(var pValue) {
    ${_table._modelLowerCase}FB._addedBlocks = setCriteria(
        pValue,
        ${_table._modelLowerCase}FB.parameters,
        param,
        SqlSyntax.IN.replaceAll(SqlSyntax.notKeyword, _waitingNot),
        ${_table._modelLowerCase}FB._addedBlocks);
    _waitingNot = '';
    ${_table._modelLowerCase}FB._addedBlocks.needEndBlock[${_table._modelLowerCase}FB._blockIndex] =
        ${_table._modelLowerCase}FB._addedBlocks.retVal;
    return ${_table._modelLowerCase}FB;
  }
}
// endregion ${_table.modelName}Field
''';
    return toString;
  }
}

class SqfEntityObjectFilterBuilder {
  SqfEntityObjectFilterBuilder(SqfEntityTableBase table) {
    _table = table;
  }
  SqfEntityTableBase _table;
  String get _createObjectFieldProperty => __createObjectFieldProperty();
  String get _recoveryMethodList => __recoveryMethodList();
  String get _deleteMethodList => __deleteMethodList();
  @override
  String toString() {
    final String toString = '''
       // region ${_table.modelName}FilterBuilder
class ${_table.modelName}FilterBuilder extends SearchCriteria {
  ${_table.modelName}FilterBuilder(${_table.modelName} obj) {
    whereString = '';
    qparams = QueryParams();
    parameters = List<DbParameter>();
    orderByList = List<String>();
    groupByList = List<String>();
    _addedBlocks = AddedBlocks(List<bool>(), List<bool>());
    _addedBlocks.needEndBlock.add(false);
    _addedBlocks.waitingStartBlock.add(false);
    _pagesize = 0;
    _page = 0;
    _obj = obj;
  }
  AddedBlocks _addedBlocks;
  int _blockIndex = 0;
  List<DbParameter> parameters;
  List<String> orderByList;
  ${_table.modelName} _obj;
  QueryParams qparams;
  int _pagesize;
  int _page;


  ${_table.modelName}FilterBuilder get and {
    if (parameters.isNotEmpty)
      {parameters[parameters.length - 1].wOperator = ' AND ';}
    return this;
  }

  ${_table.modelName}FilterBuilder get or {
    if (parameters.isNotEmpty)
      {parameters[parameters.length - 1].wOperator = ' OR ';}
    return this;
  }

  ${_table.modelName}FilterBuilder get startBlock {
    _addedBlocks.waitingStartBlock.add(true);
    _addedBlocks.needEndBlock.add(false);
    _blockIndex++;
    if (_blockIndex > 1) _addedBlocks.needEndBlock[_blockIndex - 1] = true;
    return this;
  }

  ${_table.modelName}FilterBuilder where(String whereCriteria) {
    if (whereCriteria != null && whereCriteria != '') {
      final DbParameter param = DbParameter();
      _addedBlocks = setCriteria(
          0, parameters, param, '(\$whereCriteria)', _addedBlocks);
      _addedBlocks.needEndBlock[_blockIndex] = _addedBlocks.retVal;
    }
    return this;
  }

  ${_table.modelName}FilterBuilder page(int page, int pagesize) {
    if (page > 0) _page = page;
    if (pagesize > 0) _pagesize = pagesize;
    return this;
  }

  ${_table.modelName}FilterBuilder top(int count) {
    if (count > 0) {
      _pagesize = count;
    }
    return this;
  }

  ${_table.modelName}FilterBuilder get endBlock {
    if (_addedBlocks.needEndBlock[_blockIndex]) {
      parameters[parameters.length - 1].whereString += ' ) ';
    }
    _addedBlocks.needEndBlock.removeAt(_blockIndex);
    _addedBlocks.waitingStartBlock.removeAt(_blockIndex);
    _blockIndex--;
    return this;
  }

  ${_table.modelName}FilterBuilder orderBy(var argFields) {
    if (argFields != null) {
      if (argFields is String) {
        orderByList.add(argFields); }
      else {
        for (String s in argFields) {
          if (s != null && s != '') orderByList.add(' \$s ');
        } }
    }
    return this;
  }

  ${_table.modelName}FilterBuilder orderByDesc(var argFields) {
    if (argFields != null) {
      if (argFields is String) {
        orderByList.add('\$argFields desc '); }
      else {
        for (String s in argFields) {
          if (s != null && s != '') orderByList.add(' \$s desc ');
        } }
    }
    return this;
  }

  ${_table.modelName}FilterBuilder groupBy(var argFields) {
    if (argFields != null) {
      if (argFields is String) {
        groupByList.add(' \$argFields '); }
      else {
        for (String s in argFields) {
          if (s != null && s != '') groupByList.add(' \$s ');
        } }
    }
    return this;
  }

  ${_table.modelName}Field setField(${_table.modelName}Field field, String colName, DbType dbtype) {
    return ${_table.modelName}Field(this)
    ..param = DbParameter(
        dbType: dbtype,
        columnName: colName,
        wStartBlock: _addedBlocks.waitingStartBlock[_blockIndex]);
  }

  $_createObjectFieldProperty

  bool _getIsDeleted;

  void _buildParameters() {
     if (_page > 0 && _pagesize > 0) {
      qparams
      ..limit = _pagesize
      ..offset = (_page - 1) * _pagesize;
    } else {
      qparams
      ..limit = _pagesize
      ..offset = _page;
    }
    for (DbParameter param in parameters) {
      if (param.columnName != null) {
        if (param.value is List) {
          param.value = param.value
              .toString()
              .replaceAll('[', '')
              .replaceAll(']', '')
              .toString();
          whereString += param.whereString
              .replaceAll('{field}', param.columnName)
              .replaceAll('?', param.value.toString());
          param.value = null;
        } else {
          whereString +=
              param.whereString.replaceAll('{field}', param.columnName); }
        switch (param.dbType) {
          case DbType.bool:
            if (param.value != null) param.value = param.value == true ? 1 : 0;
            break;
          default:
        }

        if (param.value != null) whereArguments.add(param.value);
        if (param.value2 != null) whereArguments.add(param.value2);
      } else {
        whereString += param.whereString;}
    }
    if (${_table.modelName}._softDeleteActivated) {
      if (whereString != '') {
        whereString = '\${!_getIsDeleted ? 'ifnull(isDeleted,0)=0 AND' : ''} (\$whereString)';}
      else if (!_getIsDeleted) {whereString = 'ifnull(isDeleted,0)=0';}
    }

    if (whereString != '') {qparams.whereString = whereString;}
        qparams
    ..whereArguments = whereArguments
    ..groupBy = groupByList.join(',')
    ..orderBy = orderByList.join(',');
  }

  
/// <summary>
/// Deletes List<${_table.modelName}> batch by query 
/// </summary>
/// <returns>BoolResult res.success=Deleted, not res.success=Can not deleted</returns>
Future<BoolResult> delete([bool hardDelete=false]) async {
  _buildParameters();
  $_deleteMethodList
    if(${_table.modelName}._softDeleteActivated && !hardDelete) {
      r = await _obj._mn${_table.modelName}.updateBatch(qparams,{'isDeleted':1}); }
  else {
      r = await _obj._mn${_table.modelName}.delete(qparams); }
  return r;    
}
  $_recoveryMethodList

  Future<BoolResult> update(Map<String, dynamic> values) {
    _buildParameters();
    return _obj._mn${_table.modelName}.updateBatch(qparams, values);
  }

  /// This method always returns ${_table.modelName}Obj if exist, otherwise returns null 
  /// <returns>List<${_table.modelName}></returns>
  Future<${_table.modelName}> toSingle([VoidCallback ${_table._modelLowerCase}(${_table.modelName} o)]) async{
    _pagesize = 1;
    _buildParameters();
    final objFuture = _obj._mn${_table.modelName}.toList(qparams);
    final data = await objFuture;
    ${_table.modelName} retVal;
    if (data.isNotEmpty) { retVal = ${_table.modelName}.fromMap(data[0] as Map<String, dynamic>); } else {retVal = null;}
      if(${_table._modelLowerCase}!=null) {${_table._modelLowerCase}(retVal);}
    return retVal;
  }
  

  /// This method always returns int.
  /// <returns>int</returns>
  Future<int> toCount([VoidCallback ${_table._modelLowerCase}Count (int c)]) async {
    _buildParameters();
    qparams.selectColumns = ['COUNT(1) AS CNT'];   
    final ${toPluralLowerName(_table._modelLowerCase)}Future = await _obj._mn${_table.modelName}.toList(qparams);
    final int count = ${toPluralLowerName(_table._modelLowerCase)}Future[0]['CNT'] as int;
    if(${_table._modelLowerCase}Count != null) {${_table._modelLowerCase}Count (count);}
    return count;
  }
  
  /// This method always returns List<${_table.modelName}>. 
  /// <returns>List<${_table.modelName}></returns>
  Future<List<${_table.modelName}>> toList([VoidCallback ${_table._modelLowerCase}List (List<${_table.modelName}> o)]) async {
    final List<${_table.modelName}> ${toPluralLowerName(_table._modelLowerCase)}Data = List<${_table.modelName}>();
    final data = await toMapList();
      final int count = data.length;
      for (int i = 0; i < count; i++) {
        ${toPluralLowerName(_table._modelLowerCase)}Data.add(${_table.modelName}.fromMap(data[i] as Map<String, dynamic>));
      }
      if (${_table._modelLowerCase}List != null) ${_table._modelLowerCase}List (${toPluralName(_table._modelLowerCase)}Data);
      return ${toPluralLowerName(_table._modelLowerCase)}Data;
  }

  /// This method always returns Json String
  Future<String> toJson() async {
    final list = List<dynamic>();
    final data = await toList();
    for (var o in data) {
      list.add(o.toMap());
    }
    return json.encode(list);
  }

  /// This method always returns Json String.
  Future<String> toJsonWithChilds() async {
    final list = List<dynamic>();
    final data = await toList();
    for (var o in data) {
      list.add(await o.toMapWithChilds());
    }
    return json.encode(list);
  }

  /// This method always returns List<dynamic>.
  /// <returns>List<dynamic></returns>
  Future<List<dynamic>> toMapList() async {
    _buildParameters();
    return await _obj._mn${_table.modelName}.toList(qparams);
  } 

  /// Returns List<DropdownMenuItem<${_table.modelName}>>
  Future<List<DropdownMenuItem<${_table.modelName}>>> toDropDownMenu(
      String displayTextColumn,
      [VoidCallback dropDownMenu(List<DropdownMenuItem<${_table.modelName}>> o)]) async {
    _buildParameters();
    final ${toPluralLowerName(_table._modelLowerCase)}Future = _obj._mn${_table.modelName}.toList(qparams);

    final data = await ${toPluralLowerName(_table._modelLowerCase)}Future;
    final int count = data.length;
    final List<DropdownMenuItem<${_table.modelName}>> items = List()
    ..add(DropdownMenuItem(
      value: ${_table.modelName}(),
      child: Text('Select ${_table.modelName}'),
    ));
    for (int i = 0; i < count; i++) {
      items.add(
        DropdownMenuItem(
          value: ${_table.modelName}.fromMap(data[i] as Map<String, dynamic>),
          child: Text(data[i][displayTextColumn].toString()),
        ),
      );
    }
    if (dropDownMenu != null) {
      dropDownMenu(items);
    }
    return items;
  }

  /// Returns List<DropdownMenuItem<int>>
  Future<List<DropdownMenuItem<int>>> toDropDownMenuInt(
      String displayTextColumn,
      [VoidCallback dropDownMenu(List<DropdownMenuItem<int>> o)]) async {
    _buildParameters();
    qparams.selectColumns=['${_table.primaryKeyName}',displayTextColumn];
    final ${toPluralLowerName(_table._modelLowerCase)}Future = _obj._mn${_table.modelName}.toList(qparams);

    final data = await ${toPluralLowerName(_table._modelLowerCase)}Future;
    final int count = data.length;
    final List<DropdownMenuItem<int>> items = List()
    ..add(DropdownMenuItem(
      value: 0,
      child: Text('Select ${_table.modelName}'),
    ));
    for (int i = 0; i < count; i++) {
      items.add(
        DropdownMenuItem(
          value: data[i]['${_table.primaryKeyName}'] as int,
          child: Text(data[i][displayTextColumn].toString()),
        ),
      );
    }
    if (dropDownMenu != null) {
      dropDownMenu(items);
    }
    return items;
  }


/// This method always returns Primary Key List<int>. 
/// <returns>List<int></returns>
Future<List<int>> toListPrimaryKey([VoidCallback ${_table.primaryKeyName}List (List<int> o),
      bool buildParameters=true]) async {
  if(buildParameters) _buildParameters();
  final List<int> ${_table.primaryKeyName}Data = List<int>();
  qparams.selectColumns= ['${_table.primaryKeyName}'];
  final ${_table.primaryKeyName}Future = await _obj._mn${_table.modelName}.toList(qparams);


    final int count = ${_table.primaryKeyName}Future.length;
    for (int i = 0; i < count; i++) {
      ${_table.primaryKeyName}Data.add(${_table.primaryKeyName}Future[i]['${_table.primaryKeyName}'] as int);
    }
    if(${_table.primaryKeyName}List != null) {${_table.primaryKeyName}List (${_table.primaryKeyName}Data);}
    return ${_table.primaryKeyName}Data;

}

/// Returns List<dynamic> for selected columns. Use this method for 'groupBy' with min,max,avg.. 
/// Sample usage: (see EXAMPLE 4.2 at https://github.com/hhtokpinar/sqfEntity#group-by)  
Future<List<dynamic>> toListObject([VoidCallback listObject(List<dynamic> o)]) async {
  _buildParameters();

  final objectFuture = _obj._mn${_table.modelName}.toList(qparams);

  final List<dynamic> objectsData = List<dynamic>();
  final data = await objectFuture;
    final int count = data.length;
    for (int i = 0; i < count; i++) {
      objectsData.add(data[i]);
    }
    if (listObject != null) {listObject(objectsData);}
    return objectsData;

}

/// Returns List<String> for selected first column
/// Sample usage: await ${_table.modelName}.select(columnsToSelect: ['columnName']).toListString()
Future<List<String>> toListString([VoidCallback listString(List<String> o)]) async {
  _buildParameters();

  final objectFuture = _obj._mn${_table.modelName}.toList(qparams);

  final List<String> objectsData = List<String>();
  final data = await objectFuture;
    final int count = data.length;
    for (int i = 0; i < count; i++) {
      objectsData.add(data[i][qparams.selectColumns[0]].toString());
    }
    if (listString != null) {listString(objectsData);}
    return objectsData;

}
}
// endregion ${_table.modelName}FilterBuilder
  
       ''';

    return toString;
  }

  String __createObjectFieldProperty() {
    final retVal =
        StringBuffer('''${_table.modelName}Field _${_table.primaryKeyName};
${_table.modelName}Field get ${_table.primaryKeyName} {
return _${_table.primaryKeyName} = setField(_${_table.primaryKeyName}, '${_table.primaryKeyName}', ${DbType.integer.toString()});
}''');

    for (SqfEntityFieldType field in _table.fields) {
      retVal.writeln('''${_table.modelName}Field _${field.fieldName};
${_table.modelName}Field get ${field.fieldName} {
return _${field.fieldName} = setField(_${field.fieldName}, '${field.fieldName}', ${field.dbType});
}''');
    }
    if (_table.useSoftDeleting) {
      retVal.writeln('''${_table.modelName}Field _isDeleted;
${_table.modelName}Field get isDeleted {
return _isDeleted = setField(_isDeleted, 'isDeleted', DbType.bool);
}''');
    }

    return retVal.toString();
  }

  String __recoveryMethodList() {
    if (!_table.useSoftDeleting) {
      return '';
    }
    String retVal = '';

    for (var tableCollecion in _table.collections) {
      switch (tableCollecion.childTableField.deleteRule) {
        case DeleteRule.SET_NULL:
          // IN THIS CASE YOU CAN NOT RECOVER CHILD ROWS
          break;
        case DeleteRule.CASCADE:
          if (tableCollecion.childTable.useSoftDeleting) {
            retVal += '''
      final ${tableCollecion.childTable.tableName}idList = await toListPrimaryKey(null, false);
      await ${tableCollecion.childTable.modelName}().select(getIsDeleted: true).isDeleted.equals(true).and.${tableCollecion.childTableField.fieldName}.inValues(${tableCollecion.childTable.tableName}idList).update({'isDeleted': 0}); 
    ''';
          }
          break;
        case DeleteRule.SET_DEFAULT_VALUE:
          //IN THIS CASE YOU CAN NOT RECOVER CHILD ROWS
          break;
        default:
      }
    }

    retVal += '''
            return _obj._mn${_table.modelName}.updateBatch(qparams,{'isDeleted':0});
      ''';

    return '''
  Future<BoolResult> recover() async {
  _getIsDeleted = true;
  _buildParameters();
  print('SQFENTITIY: recover ${_table.modelName} batch invoked');
  $retVal 
  }''';
  }

  String __deleteMethodList() {
    String retVal = 'var r= BoolResult();';

    for (var tableCollecion in _table.collections) {
      switch (tableCollecion.childTableField.deleteRule) {
        case DeleteRule.SET_NULL:
          // IN THIS CASE YOU CAN NOT RECOVER CHILD ROWS
          break;
        case DeleteRule.NO_ACTION:
          retVal += '''
    var ${tableCollecion.childTable.tableName}idList = await toListPrimaryKey((_) {}, false);
    if (await ${tableCollecion.childTable.modelName}().select().${tableCollecion.childTableField.fieldName}.inValues(${tableCollecion.childTable.tableName}idList).toCount()>0) {
    return BoolResult(
    success: false,
    errorMessage: 
          'SQFENTITY ERROR: The DELETE statement conflicted with the REFERENCE RELATIONSHIP '${tableCollecion.childTable.modelName}.${tableCollecion.childTableField.fieldName}'');
    } else
                      ''';

          break;
        case DeleteRule.CASCADE:
          if (tableCollecion.childTable.useSoftDeleting) {
            retVal += '''
    final ${tableCollecion.childTable.tableName}idList = await toListPrimaryKey(null, false);
    await ${tableCollecion.childTable.modelName}().select().${tableCollecion.childTableField.fieldName}.inValues(${tableCollecion.childTable.tableName}idList).delete(hardDelete);
           ''';
          }
          break;
        case DeleteRule.SET_DEFAULT_VALUE:
          //IN THIS CASE YOU CAN NOT RECOVER CHILD ROWS
          break;
        default:
      }
    }

    return retVal;
  }
}

// END SQFENTITY BUILDERS

// BEGIN ENUMS, CLASSES AND ABSTRACTS
class TableCollectionBase {
  TableCollectionBase(this.childTable, this.childTableField);
  SqfEntityTableBase childTable;
  SqfEntityFieldRelationshipBase childTableField;
}

class TableField {
  TableField([this.fieldName, this.fieldType]);
  String fieldName;
  DbType fieldType;

  //start SQL Aggregate Functions
  String max([String alias]) {
    return 'MAX($fieldName) AS ${(alias ?? fieldName)}';
  }

  String min([String alias]) {
    return 'MIN($fieldName) AS ${(alias ?? fieldName)}';
  }

  String sum([String alias]) {
    return 'SUM($fieldName) AS ${(alias ?? fieldName)}';
  }

  String avg([String alias]) {
    return 'AVG($fieldName) AS ${(alias ?? fieldName)}';
  }

  String count([String alias]) {
    return 'COUNT($fieldName) AS ${(alias ?? fieldName)}';
  }

  //end  SQL Aggregate Functions

  // start SQL Scalar Functions

  String lTrim([String alias]) {
    return 'LTRIM($fieldName) AS ${(alias ?? fieldName)}';
  }

  String trim([String alias]) {
    return 'RTRIM(LTRIM($fieldName)) AS ${(alias ?? fieldName)}';
  }

  String rTrim([String alias]) {
    return 'RTRIM($fieldName) AS ${(alias ?? fieldName)}';
  }

  String len([String alias]) {
    return 'LEN($fieldName) AS ${(alias ?? fieldName)}';
  }

  // end SQL Scalar Functions

  String isNull(String val, [String alias]) {
    return "ifnull($fieldName,'$val') AS ${(alias ?? fieldName)}";
  }

  @override
  String toString([String alias]) {
    return '$fieldName ${alias != null ? ' AS $alias' : ''}';
  }
}

class SqlSyntax {
  static TableField setField(TableField field, String fName, [DbType fType]) {
    if (field == null) {
      field = TableField()..fieldName = fName;
      if (fType != null) {
        field.fieldType = fType;
      }
    }
    return field;
  }

  static const String notKeyword = '{isNot}';
// SQLITE keywords
// 'add','all','alter','and','as','autoincrement','between','case','check','collate','commit','constraint','create','default','deferrable','delete','distinct','drop','else','escape','except','exists','foreign','from','group','having','if','in','index','insert','intersect','into','is','isnull','join','limit','not','notnull','null','on','or','order','primary','references','select','set','table','then','to','transaction','union','unique','update','using','values','when','where'
  static const String EQuals = ' {field}=? ';
  static const String IsNULL = ' {field} IS $notKeyword NULL '; // *
  static const String NotEQuals = ' {field}<>? ';
  static const String Contains = ' {field} $notKeyword LIKE ? '; // *
  static const String Between = ' {field} $notKeyword BETWEEN ? AND ? '; // *
  static const String GreaterThan = ' {field}>? ';
  static const String LessThan = ' {field}<? ';
  static const String GreaterOrEqualsThan = ' {field}>=? ';
  static const String LessOrEqualsThan = ' {field}<=? ';
  static const String GreaterThanOrEquals = ' {field}>=? ';
  static const String LessThanOrEquals = ' {field}<=? ';
  static const String IN = ' {field} $notKeyword IN (?) '; // *
}

class QueryParams {
  QueryParams(
      {this.selectColumns,
      this.whereString,
      this.whereArguments,
      this.orderBy,
      this.groupBy,
      this.having,
      this.distinct,
      this.limit,
      this.offset});
  List<String> selectColumns;
  String whereString;
  List<dynamic> whereArguments;
  String orderBy;
  int limit;
  int offset;
  String groupBy;
  String having;
  bool distinct;
}

abstract class SearchCriteria {
  BoolResult result;
  List<String> groupByList = List<String>();
  List<dynamic> whereArguments = List<dynamic>();
  String whereString = '';

  AddedBlocks setCriteria(dynamic pValue, List<DbParameter> parameters,
      DbParameter param, String sqlSyntax, AddedBlocks addedBlocks,
      [dynamic pValue2]) {
    bool sp = addedBlocks.needEndBlock[addedBlocks.needEndBlock.length - 1];
    if (pValue != null) {
      param.whereString += parameters.isNotEmpty
          ? parameters[parameters.length - 1].wOperator
          : '';

      for (int i = 1; i < addedBlocks.waitingStartBlock.length; i++) {
        if (addedBlocks.waitingStartBlock[i]) {
          param.whereString += ' ( ';
          addedBlocks.waitingStartBlock[i] = false;
          sp = true;
        }
      }

      param.value = pValue;
      if (pValue2 != null) {
        param.value2 = pValue2;
      }
      param.whereString += sqlSyntax;
      parameters.add(param);
    }
    addedBlocks.retVal = sp;

    return addedBlocks;
  }
}

class AddedBlocks {
  AddedBlocks([this.needEndBlock, this.waitingStartBlock, this.retVal = false]);
  bool retVal;
  List<bool> needEndBlock;
  List<bool> waitingStartBlock;
}

class BoolResult {
  BoolResult({this.success, this.successMessage, this.errorMessage});
  String successMessage;
  String errorMessage;
  bool success = false;

  @override
  String toString() {
    if (success) {
      return successMessage != null && successMessage != ''
          ? successMessage
          : 'Result: OK! Successful';
    } else {
      return errorMessage != null && errorMessage != ''
          ? errorMessage
          : 'Result: ERROR!';
    }
  }
}

class DbParameter {
  DbParameter(
      {this.columnName,
      this.dbType,
      this.expression = '',
      this.value,
      this.value2,
      this.whereString = '',
      this.wOperator = '',
      this.wStartBlock = false});
  String columnName;
  DbType dbType;
  dynamic value;
  dynamic value2;
  String whereString;
  bool wStartBlock;
  String wOperator;
  String expression;
}

class SqfEntityTableBase {
  String tableName;
  String primaryKeyName;
  List<SqfEntityFieldType> fields;
  List<TableCollectionBase> collections;
  bool useSoftDeleting = false;
  //bool primaryKeyisIdentity = true;
  PrimaryKeyType primaryKeyType = PrimaryKeyType.integer_auto_incremental;
  String modelName;
  String dbModel;
  String defaultJsonUrl;
  bool initialized;

  SqfEntityTableBase init() {
    modelName = toModelName(modelName, tableName);
    primaryKeyType = primaryKeyType ?? PrimaryKeyType.integer_auto_incremental;
    initialized = false;
    // print( '>>>>>>>>>>>>>>>>>>>>>>>>>>>> SqfEntityTableBase of [$tableName] initialized successfuly');
    return this;
  }

  String get _modelLowerCase => modelName.toLowerCase();

  String get createTableSQL => _createTableSQL();
  String _createTableSQL() {
    final _createTableSQL = StringBuffer('');
    switch (primaryKeyType) {
      case PrimaryKeyType.integer_unique:
        _createTableSQL.write('int UNIQUE');
        break;
      case PrimaryKeyType.text:
        _createTableSQL.write('text UNIQUE');
        break;
      default:
        _createTableSQL.write('integer primary key');
    }
    for (SqfEntityFieldType field in fields) {
      _createTableSQL.write(', ${field.toSqLiteFieldString()}');
    }
    if (useSoftDeleting) {
      _createTableSQL.write(', isDeleted numeric');
    }
    return 'Create table $tableName ($primaryKeyName ${_createTableSQL.toString()})';
  }
}

class SqfEntitySequenceBase {
  /// Name of the sequence.
  String sequenceName;

  /// starting value from where the sequence starts. startWith value should be greater than or equal
  /// to minimum value and less than equal to maximum value.
  int startWith = 0;

  ///Value by which sequence will increment itself. Increment_value can be positive or negative.
  int incrementBy = 1;

  /// Minimum value of the sequence.
  int minValue = 0;

  /// Maximum value of the sequence.
  int maxValue = 9223372036854775807;

  /// cycle if true then, When sequence reaches its max value it starts from beginning. otherwise An exception will be thrown if sequence exceeds its max_value.
  bool cycle = false;

  String modelName;

  bool initialized = false;
  SqfEntitySequenceBase init() {
    startWith = startWith ?? 0;
    incrementBy = incrementBy ?? 1;
    minValue = minValue ?? 0;
    maxValue = maxValue ?? 9223372036854775807;
    cycle = cycle ?? false;
    if (minValue >= maxValue) {
      throw Exception(
          'SQFENTITIY: SEQUENCE ($sequenceName) INITIALIZE ERROR! The maxValue must be greater than minValue');
    } else if (!(minValue <= startWith && startWith < maxValue)) {
      throw Exception(
          'SQFENTITIY: SEQUENCE ($sequenceName) INITIALIZE ERROR! The startWith value must between minValue and maxValue');
    }

    modelName = modelName ??
        '${sequenceName.substring(0, 1).toUpperCase() + sequenceName.substring(1).toLowerCase()}Sequence';

    // print('>>>>>>>>>>>>>>>>>>>>>>>>>>>> SqfEntitySequenceBase of [$sequenceName] initialized successfuly');
    return this;
  }
}

abstract class SqfEntityFieldType {
  SqfEntityFieldType(this.fieldName, this.dbType,
      {this.defaultValue, this.table, this.sequencedBy});
  final String fieldName;
  DbType dbType;
  String _dbType;
  dynamic defaultValue;
  SqfEntityTableBase table;
  SqfEntitySequenceBase sequencedBy;
  String toSqLiteFieldString();

  String toPropertiesString();

  String toConstructureString();

  String toMapString();

  String toFromMapString();
}

class SqfEntityFieldBase implements SqfEntityFieldType {
  @override
  SqfEntityFieldBase(this.fieldName, this.dbType,
      {this.defaultValue, this.table, this.sequencedBy});
  @override
  String fieldName;
  @override
  DbType dbType;
  @override
  String _dbType;
  @override
  dynamic defaultValue;
  @override
  SqfEntityTableBase table;
  @override
  SqfEntitySequenceBase sequencedBy;
  @override
  String toSqLiteFieldString() {
    switch (dbType) {
      case DbType.bool:
        _dbType = 'numeric';
        break;
      default:
        _dbType = dbType.toString().replaceAll('DbType.', '');
    }
    return '$fieldName $_dbType';
  }

  @override
  String toPropertiesString() {
    return '${dartType[dbType.index]} $fieldName';
  }

  @override
  String toConstructureString() {
    return 'this.$fieldName';
  }

  @override
  String toMapString() {
    switch (dbType) {
      case DbType.bool: //forQuery? (bxcol8Bool? 1 : 0):bxcol8Bool;
        {
          return 'if ($fieldName != null) {map[\'$fieldName\'] = forQuery? ($fieldName ? 1 : 0) : $fieldName;}\n';
        }
        break;
      default:
        {
          return 'if ($fieldName != null) {map[\'$fieldName\'] = $fieldName;}\n';
        }
    }
  }

  @override
  String toFromMapString() {
    if (dbType == DbType.bool) {
      return '$fieldName = o[\'$fieldName\'] != null ? o[\'$fieldName\'] == 1 : null;\n';
    } else {
      return '$fieldName = o[\'$fieldName\'] as ${dartType[dbType.index].toString()};\n';
    }
  }
}

class SqfEntityFieldRelationshipBase implements SqfEntityFieldType {
  SqfEntityFieldRelationshipBase(this.table, this.deleteRule,
      {this.fieldName, this.defaultValue}) {
    fieldName = fieldName ??
        table.tableName +
            table.primaryKeyName.substring(0, 1).toUpperCase() +
            table.primaryKeyName.substring(1);

    relationshipName = table.modelName;
  }
  @override
  String fieldName;
  String relationshipName;
  @override
  DbType dbType = DbType.integer;
  @override
  String _dbType;
  @override
  dynamic defaultValue;
  @override
  SqfEntityTableBase table;
  @override
  SqfEntitySequenceBase sequencedBy;
  DeleteRule deleteRule = DeleteRule.NO_ACTION;

  @override
  String toSqLiteFieldString() {
    _dbType = 'integer';
    return '$fieldName $_dbType';
  }

  @override
  String toPropertiesString() {
    return 'int $fieldName';
  }

  @override
  String toConstructureString() {
    return 'this.$fieldName';
  }

  @override
  String toMapString() {
    switch (dbType) {
      case DbType.bool: //forQuery? (bxcol8Bool? 1 : 0):bxcol8Bool;
        return 'if ($fieldName != null) {map[\'$fieldName\'] = forQuery? ($fieldName ? 1 : 0) : $fieldName;}\n';
      default:
        return 'if ($fieldName != null) {map[\'$fieldName\'] = $fieldName;}\n';
    }
  }

  @override
  String toFromMapString() {
    switch (dbType) {
      case DbType.bool:
        return '$fieldName = o[\'$fieldName\'] != null ? o[\'$fieldName\'] == 1 : null;\n';
      default:
        return '$fieldName = o[\'$fieldName\'] as ${dartType[dbType.index].toString()};\n';
    }
  }
}

abstract class SqfEntityModelBase {
    String modelName;
  String databaseName;
  String bundledDatabasePath;
  List<SqfEntityTableBase> databaseTables;
  List<SqfEntitySequenceBase> sequences;

}

String toCamelCase(String fieldName) => fieldName.length == 1
    ? fieldName.toUpperCase()
    : fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);

String tocamelCase(String fieldName) => fieldName.length == 1
    ? fieldName.toLowerCase()
    : fieldName.substring(0, 1).toLowerCase() + fieldName.substring(1);


String toPluralName(String s) => s.endsWith('y')
    ? '${s.substring(0, s.length - 1)}ies'
    : (s.endsWith('s')) ? '${s}es' : '${s}s';

String toPluralLowerName(String s) => s.endsWith('y')
    ? '${s.substring(0, s.length - 1).toLowerCase()}ies'
    : (s.endsWith('s')) ? '${s.toLowerCase()}es' : '${s.toLowerCase()}s';

String toSingularName(String s) => s.endsWith('ies')
    ? '${s.substring(0, s.length - 3)}y'
    : s.endsWith('ses')
        ? '${s.substring(0, s.length - 3)}s'
        : s.endsWith('s') ? s.substring(0, s.length - 1) : s;

String toSingularLowerName(String s) => s.endsWith('ies')
    ? '${s.substring(0, s.length - 3).toLowerCase()}y'
    : s.endsWith('ses')
        ? '${s.substring(0, s.length - 3).toLowerCase()}s'
        : s.endsWith('s')
            ? s.substring(0, s.length - 1).toLowerCase()
            : s.toLowerCase();

String toModelName(String modelName, String definedName) => modelName == null
    ? toCamelCase(toSingularName(definedName))
    : toCamelCase(modelName);
/*
SQFLITE NOTES:
Supported SQLite types
No validity check is done on values yet so please avoid non supported types https://www.sqlite.org/datatype3.html

DateTime is not a supported SQLite type. Personally I store them as int (millisSinceEpoch) or string (iso8601)

bool is not a supported SQLite type. Use INTEGER and 0 and 1 values.

Each value stored in an SQLite database (or manipulated by the database engine) has one of the following storage classes:

-- NULL    -> The value is a NULL value.
-- INTEGER -> The value is a signed integer, stored in 1, 2, 3, 4, 6, or 8 bytes depending on the magnitude of the value.
-- REAL    -> The value is a floating point value, stored as an 8-byte IEEE floating point number.
-- TEXT    -> The value is a text string, stored using the database encoding (UTF-8, UTF-16BE or UTF-16LE).
-- BLOB    -> The value is a blob of data, stored exactly as it was input.


SQLite KEYWORDS

'add','all','alter','and','as','autoincrement',
'between','case','check','collate','commit','constraint','create',
'default','deferrable','delete','distinct','drop',
'else','escape','except','exists',
'foreign','from','group','having',
'if','in','index','insert','intersect','into','is','isnull',
'join','limit','not','notnull','null','on','or','order',
'primary','references','select','set',
'table','then','to','transaction',
'union','unique','update','using',
'values','when','where'

*/

// List<String> dartType: each element of the list corresponds to the DbType enum index
const List<String> dartType = [
  'int',
  'String',
  'String',
  'double',
  'int',
  'bool'
];
const List<String> sqLiteType = [
  'integer',
  'text',
  'blob',
  'real',
  'numeric',
  'numeric'
];
enum DeleteRule { CASCADE, SET_DEFAULT_VALUE, SET_NULL, NO_ACTION }
enum PrimaryKeyType { integer_auto_incremental, text, integer_unique }
enum DbType {
  integer,
// ------------------------------------------------------
// INT                             |
// INTEGER                         |
// TINYINT                         |
// SMALLINT                        |
// MEDIUMINT                       |
// BIGINT                          |        INTEGER
// UNSIGNED BIG INT                |
// INT2                            |
// INT8	                           |
// -----------------------------------------------------
  text,
// -----------------------------------------------------
// CHARACTER(20)                   |
// VARCHAR(255)                    |
// VARYING CHARACTER(255)          |
// NCHAR(55)                       |
// NATIVE CHARACTER(70)            |        TEXT
// NVARCHAR(100)                   |
// TEXT                            |
// CLOB                            |
// ------------------------------------------------------
  blob,
// ------------------------------------------------------
// BLOB                            |        BLOB
// ------------------------------------------------------
  real,
// ------------------------------------------------------
// REAL                            |
// DOUBLE                          |
// DOUBLE PRECISION                |        REAL
// FLOAT	                         |
// ------------------------------------------------------
  numeric,
// ------------------------------------------------------
// NUMERIC                         |
// DECIMAL(10,5)                   |
// BOOLEAN                         |       NUMERIC
// DATE                            |
// DATETIME	                       |
// -------------------------------------------------------
  bool
  // bool is not a supported SQLite type. Builder converts this type to numeric values (false=0, true=1)
}

DbType parseDbType(String val) {
  for (var o in DbType.values) {
    if (sqLiteType[o.index] == val) {
      return o;
    }
  }
  return null;
}
// END ENUMS, CLASSES AND ABSTRACTS


class _DbModel extends SqfEntityModelBase {}


